@@ -3,6 +3,7 @@ import re
 import json
 import time
 import logging
+from typing import Dict
 
 from smda.common.SmdaReport import SmdaReport
 
@@ -10,9 +11,12 @@ from mcrit.config.McritConfig import McritConfig
 from mcrit.config.MinHashConfig import MinHashConfig
 from mcrit.config.ShinglerConfig import ShinglerConfig
 from mcrit.config.StorageConfig import StorageConfig
+from mcrit.index.SearchCursor import MinimalSearchCursor, FullSearchCursor
+from mcrit.index.SearchQueryParser import SearchQueryParser
 from mcrit.libs.utility import compress_encode, decompress_decode
 from mcrit.queue.QueueFactory import QueueFactory
 from mcrit.queue.QueueRemoteCalls import QueueRemoteCaller, NoProgressReporter
+from mcrit.storage.FamilyEntry import FamilyEntry
 from mcrit.storage.FunctionEntry import FunctionEntry
 from mcrit.storage.SampleEntry import SampleEntry
 from mcrit.storage.StorageFactory import StorageFactory
@@ -23,6 +27,31 @@ logging.basicConfig(level=logging.INFO)
 LOGGER = logging.getLogger(__name__)
 
 
+#### Helper methods for nested sort_by in search
+
+def _get_field_once(object, field):
+    try:
+        return getattr(object, field)
+    except Exception:
+        pass
+    try:
+        return object[field]
+    except Exception:
+        pass
+
+def _get_field(object, field):
+    dot_index = field.find(".")
+    if dot_index >= 0:
+        first_field = field[:dot_index]
+        remaining_fields = field[dot_index+1:]
+        return _get_field(
+            _get_field_once(object, first_field),
+            remaining_fields
+        )
+    else:
+        return _get_field_once(object, field)
+
+
 class MinHashIndex(QueueRemoteCaller(Worker)):
     def __init__(self, config=None, base_url=None):
         if config is None:
@@ -34,8 +63,9 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
         self._storage = StorageFactory.getStorage(config.STORAGE_CONFIG)
         # config.QUEUE_CONFIG.QUEUE_METHOD = QueueFactory.QUEUE_METHOD_FAKE
         queue = QueueFactory().getQueue(config, storage=self._storage, consumer_id="index")
+        self.search_query_parser = SearchQueryParser()
         super().__init__(queue)
-
+    
     #### STORAGE IO ####
     def getStorage(self):
         """Get an interface to the storage"""
@@ -116,6 +146,7 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
             "num_families_skipped": 0
         }
         # TODO adjust minimum required MCRIT version if we ever extend in a way that objects become incompatible
+        # TODO do not compare version as string
         if export_data["config"]["version"] <= "0.0.0":
             LOGGER.error("Cannot import data, version of export is incompatible / too old.")
             return 
@@ -138,9 +169,15 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
             else:
                 import_report["num_families_skipped"] += 1
             family_id_remapping[exported_family_id] = remapped_family_id
-        LOGGER.info("Family remapping created...")
+        LOGGER.info("Family remapping created: %d families, %d samples.", 
+                len(family_id_remapping), 
+                len(export_data["sample_entries"]))
         # iterate samples
+        index = 0
+        function_entries_to_import = []
+        minhashes_to_import = []
         for sample_sha256, sample_entry_dict in export_data["sample_entries"].items():
+            index += 1
             # check if sample is already present, and skip if yes
             if self._storage.getSampleBySha256(sample_sha256):
                 LOGGER.info(f"Sample with SHA256 {sample_sha256} already present in database, skipping...")
@@ -159,20 +196,23 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
                 if is_compressed:
                     function_entries = json.loads(decompress_decode(function_entries))
                 # iterate functions and add them, adjusting family_id and sample_id
-                minhashes = []
                 for old_function_id, function_entry_dict in function_entries.items():
                     function_entry = FunctionEntry.fromDict(function_entry_dict)
                     function_entry.sample_id = remapped_sample_entry.sample_id
                     function_entry.family_id = remapped_sample_entry.family_id
-                    remapped_function_entry = self._storage.importFunctionEntry(function_entry)
+                    function_entries_to_import.append(function_entry)
                     # delaying minhash insertion speeds up the procedure dramatically
                     if function_entry.minhash:
                         minhash = MinHash(function_entry.function_id, function_entry.minhash, minhash_bits=self._minhash_config.MINHASH_SIGNATURE_BITS)
-                        minhashes.append(minhash)
+                        minhashes_to_import.append(minhash)
                     import_report["num_functions_imported"] += 1
-                # ensure that their minhashes / pichashes are added to the respective indices
-                self._storage.addMinHashes(minhashes)
-            LOGGER.info(f"Sample with SHA256 {sample_sha256} added...")
+            LOGGER.info(f"Sample %d with SHA256 %s added...", index, sample_sha256)
+        
+        self._storage.importFunctionEntries(function_entries_to_import)
+        LOGGER.info(f"{len(function_entries_to_import)} functions added")
+        # ensure that their minhashes / pichashes are added to the respective indices
+        self._storage.addMinHashes(minhashes_to_import)
+        LOGGER.info(f"{len(minhashes_to_import)} Minhashes added")
         return import_report
 
     def respawn(self):
@@ -197,7 +237,7 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
     def addReport(self, smda_report, calculate_hashes=True, calculate_matches=False):
         sample_entry = self._storage.getSampleBySha256(smda_report.sha256)
         if sample_entry:
-            return {"sample_info": sample_entry.toDict()}
+            return {"existed": True, "sample_info": sample_entry.toDict()}
         sample_entry = self._storage.addSmdaReport(smda_report)
         if not sample_entry:
             return None
@@ -207,7 +247,7 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
         job_id = None
         if calculate_hashes:
             job_id = self.updateMinHashesForSample(sample_entry.sample_id)
-        return {"sample_info": sample_entry.toDict(), "job_id": job_id}
+        return {"existed": False, "sample_info": sample_entry.toDict(), "job_id": job_id}
 
     def addReportJson(self, report_json, calculate_hashes=True, calculate_matches=False):
         report = SmdaReport.fromDict(report_json)
@@ -235,8 +275,11 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
     def deleteSample(self, sample_id):
         return self._storage.deleteSample(sample_id)
 
-    def getFunctionById(self, function_id):
-        return self._storage.getFunctionById(function_id)
+    def deleteFamily(self, family_id, keep_samples=False):
+        return self._storage.deleteFamily(family_id, keep_samples=keep_samples)
+
+    def getFunctionById(self, function_id, with_xcfg=False):
+        return self._storage.getFunctionById(function_id, with_xcfg=with_xcfg)
 
     def getFunctions(self, start_index, limit):
         return self._storage.getFunctions(start_index, limit)
@@ -250,26 +293,10 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
     def getSamples(self, start_index, limit):
         return self._storage.getSamples(start_index, limit)
 
-    def getFamilies(self):
+    def getFamilies(self) -> Dict[int, FamilyEntry]:
         family_overview = {}
         for family_id in self._storage.getFamilyIds():
-            family_is_library = False
-            overview_entry = {
-                "family_id": family_id,
-                "family": self._storage.getFamily(family_id),
-                "num_samples": 0,
-                "num_functions": 0,
-                "is_library": family_is_library
-            }
-            sample_entries = self._storage.getSamplesByFamilyId(family_id)
-            if sample_entries:
-                family_is_library = True
-            for sample_entry in sample_entries:
-                overview_entry["num_samples"] += 1
-                overview_entry["num_functions"] += sample_entry.statistics["num_functions"]
-                family_is_library &= sample_entry.is_library
-            overview_entry["is_library"] = family_is_library
-            family_overview[family_id] = overview_entry
+            family_overview[family_id] = self._storage.getFamily(family_id)
         return family_overview
 
     def getFunctionGraph(self, function_id):
@@ -289,6 +316,7 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
         storage_stats = self._storage.getStats()
         status = {
             "status": {
+                "db_state": storage_stats["db_state"],
                 "storage_type": self._storage_config.STORAGE_METHOD,
                 "storage_bands": len(storage_stats["bands"]),
                 "num_samples": storage_stats["num_samples"],
@@ -299,30 +327,105 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
         }
         return status
 
-    def getSearchResults(self, search_term, max_num_results=100):
-        """ Recognize characteristics of search term and conduct applicable searches across compatible fields """
-        results = {
-            "families": {},
-            "samples": {},
-            "functions": {},
-            # TODO decide whether jobs/results makes more sense here
-            "jobs": {},
-            "stats": {
-                "max_num_results": max_num_results,
-                "num_families_found": 0,
-                "num_samples_found": 0,
-                "num_functions_found": 0,
-                "num_jobs_found": 0,
-                }
+    def getMatchesForPicHash(self, pichash):
+        return self._storage.getMatchesForPicHash(pichash)
+        
+    def getMatchesForPicBlockHash(self, picblockhash):
+        return self._storage.getMatchesForPicBlockHash(picblockhash)
+
+    def getSampleBySha256(self, sample_sha256): 
+        return self._storage.getSampleBySha256(sample_sha256)
+
+    #### SEARCH ####
+
+    # When performing an initial search, the cursor should be set to None.
+    # Search results are of the following form:
+    # {
+    #     "search_results": {
+    #         id1: found_entry1,
+    #         id2: found_entry2,
+    #         ...
+    #     },
+    #     "cursor": {
+    #         "forward": forward cursor,
+    #         "backward": backward cursor,
+    #     } 
+    # }
+    # To get further results, perform a search using the forward cursor.
+    # To get back to the previous search results, use the backward cursor.
+    # If no further or previous results are available, the forward or backward cursor will be None.
+    #
+    # IMPORTANT: A cursor shall only be used in combination with the same
+    # search_term and sort_by_list that were used when the cursor was returned from mcrit.
+    # If those parameters are altered, mcrit's behavior is undefined.
+
+    def _getSearchResultTemplate(self, search_function, search_term, sort_by_list, cursor_str, limit):
+        assert isinstance(search_term, str)
+
+        sort_fields = [sort_info[0] for sort_info in sort_by_list]
+
+        is_backward_search = False
+        cursor = MinimalSearchCursor.fromStr(cursor_str)
+
+        full_cursor = FullSearchCursor(cursor, sort_by_list)
+        is_backward_search = not full_cursor.is_forward_search
+
+        parsed_search = self.search_query_parser.parse(search_term)
+        search_results_objects = search_function(parsed_search, cursor=full_cursor, max_num_results=limit+1)
+
+        # Find last last_element_key, which is used for the forward cursor
+        search_results_keys = list(search_results_objects.keys())
+        if len(search_results_objects) >= limit + 1:
+            search_results_objects.pop(search_results_keys[-1])
+            search_results_keys.pop()
+            last_element_key = search_results_keys[-1]
+        else:
+            last_element_key = None
+        
+        forward_cursor_str = None
+        if last_element_key:
+            last_result = search_results_objects[last_element_key]
+            forward_cursor = MinimalSearchCursor()
+            forward_cursor.is_forward_search = True ^ is_backward_search # switch for backward search, because of swap
+            forward_cursor.record_values = [_get_field(last_result, field) for field in sort_fields]
+            forward_cursor_str = forward_cursor.toStr()
+
+        backward_cursor_str = None
+        if len(search_results_objects) > 0 and cursor is not None:
+            first_result = search_results_objects[search_results_keys[0]]
+            backward_cursor = MinimalSearchCursor()
+            backward_cursor.is_forward_search = False ^ is_backward_search # switch for backward search, because of swap
+            backward_cursor.record_values = [_get_field(first_result, field) for field in sort_fields]
+            backward_cursor_str = backward_cursor.toStr()
+
+        if is_backward_search:
+            # reverse order
+            backward_cursor_str, forward_cursor_str = forward_cursor_str, backward_cursor_str
+            search_results = {k: v.toDict() for k, v in reversed(search_results_objects.items())}
+        else:
+            search_results = {k: v.toDict() for k, v in search_results_objects.items()}
+
+        return {
+            "search_results": search_results,
+            "cursor": {
+                "forward": forward_cursor_str,
+                "backward": backward_cursor_str,
+            }
         }
-        # try to regex as sha256 hash: sample_id
-        if re.match("^[a-fA-F0-9]{64}$", search_term) is not None:
-            sample_entry = self._storage.getSampleBySha256(search_term)
-            results["samples"][sample_entry.sample_id] = sample_entry.toDict()
-            results["stats"]["num_samples_found"] = 1
-            return results
-        # try to parse as int: family_id, sample_id, function_id, function_address, pic_hash
+
+    def _get_sort_data(self, standard_sort, sort_by, is_ascending):
+        if sort_by == standard_sort or sort_by is None:
+            sort_by_list = [(standard_sort, is_ascending),]
+        else:
+            sort_by_list = [
+                (sort_by, is_ascending),
+                (standard_sort, True),
+            ]
+        return sort_by_list 
+
+    def getFamilySearchResults(self, search_term, sort_by="family_id", is_ascending=True, cursor=None, limit=100):
         term_as_int = None
+        id_match = None
         try:
             if search_term.startswith("0x"):
                 term_as_int = int(search_term, 16)
@@ -330,35 +433,125 @@ class MinHashIndex(QueueRemoteCaller(Worker)):
                 term_as_int = int(search_term)
             if term_as_int <= 0xFFFFFFFF:
                 if self._storage.isFamilyId(term_as_int):
-                    results["families"][term_as_int] = self._storage.getFamily(term_as_int)
-                    results["stats"]["num_families_found"] = 1
-                if self._storage.isSampleId(term_as_int):
-                    results["samples"][term_as_int] = self._storage.getSampleById(term_as_int).toDict()
-                    results["stats"]["num_samples_found"] = 1
-                if self._storage.isFunctionId(term_as_int):
-                    results["functions"][term_as_int] = self._storage.getFunctionById(term_as_int).toDict()
-                    results["stats"]["num_functions_found"] = 1
+                    id_match = self._storage.getFamily(term_as_int).toDict()
             else:
                 LOGGER.warn("Can only handle family/sample/function IDs up to 0xFFFFFFFF.")
-            if self._storage.isPicHash(term_as_int):
-                pic_matches = self._storage.getMatchesForPicHash(term_as_int)
-                results["stats"]["num_functions_found"] += len(pic_matches)
-                for match in pic_matches:
-                    sample_id, function_id = match
-                    if len(results["functions"]) < max_num_results:
-                        results["functions"][function_id] = self._storage.getFunctionById(function_id).toDict()
-            return results
-        except:
+        except Exception:
             pass
-        # as regular string, refer to storage implementation of searching
-        results["families"].update(self._storage.findFamilyByString(search_term, max_num_results=max_num_results))
-        results["samples"].update({k: v.toDict() for k, v in self._storage.findSampleByString(search_term, max_num_results=max_num_results).items()})
-        results["functions"].update({k: v.toDict() for k, v in self._storage.findFunctionByString(search_term, max_num_results=max_num_results).items()})
-        # NOTE right now, we cap the count to 100 due to Storage implementation, could also count all and just limit returned results
-        results["stats"]["num_families_found"] = len(results["families"])
-        results["stats"]["num_samples_found"] = len(results["samples"])
-        results["stats"]["num_functions_found"] = len(results["functions"])
-        return results
+
+        assert sort_by in (
+            None,
+            "family_id",
+            "family_name",
+            "num_samples",
+            "num_library_samples",
+            "num_functions",
+        )
+
+        sort_data = self._get_sort_data("family_id", sort_by, is_ascending)
+        result = self._getSearchResultTemplate(
+            self._storage.findFamilyByString,
+            search_term,
+            sort_data,
+            cursor,
+            limit,
+        )
+        result["id_match"] = id_match
+        return result
+
+    def getFunctionSearchResults(self, search_term, sort_by="function_id", is_ascending=True, cursor=None, limit=100):
+        term_as_int = None
+        id_match = None
+        try:
+            if search_term.startswith("0x"):
+                term_as_int = int(search_term, 16)
+            else:
+                term_as_int = int(search_term)
+            if term_as_int <= 0xFFFFFFFF:
+                if self._storage.isFunctionId(term_as_int):
+                    id_match = self._storage.getFunctionById(term_as_int).toDict()
+            else:
+                LOGGER.warn("Can only handle family/sample/function IDs up to 0xFFFFFFFF.")
+        except Exception:
+            pass
+
+        assert sort_by in (
+            None,
+            "function_id",
+            "sample_id",
+            "family_id",
+            "pichash",
+            "function_name",
+            "offset",
+            "num_instructions",
+            "num_blocks",
+        )
+        
+        sort_data = self._get_sort_data("function_id", sort_by, is_ascending)
+        result = self._getSearchResultTemplate(
+            self._storage.findFunctionByString,
+            search_term,
+            sort_data,
+            cursor,
+            limit,
+        )
+        result["id_match"] = id_match
+        return result
+
+    def getSampleSearchResults(self, search_term, sort_by="sample_id", is_ascending=True, cursor=None, limit=100):
+        term_as_int = None
+        id_match = None
+        try:
+            if search_term.startswith("0x"):
+                term_as_int = int(search_term, 16)
+            else:
+                term_as_int = int(search_term)
+            if term_as_int <= 0xFFFFFFFF:
+                if self._storage.isSampleId(term_as_int):
+                    id_match = self._storage.getSampleById(term_as_int).toDict()
+            else:
+                LOGGER.warn("Can only handle family/sample/function IDs up to 0xFFFFFFFF.")
+        except Exception:
+            pass
+
+        if re.match("^[a-fA-F0-9]{64}$", search_term) is not None:
+            sample_entry = self._storage.getSampleBySha256(search_term)
+            sha_match = sample_entry.toDict()
+        else:
+            sha_match = None
+
+        assert sort_by in (
+            None,
+            "filename",
+            "function_id",
+            "sample_id",
+            "family_id",
+            "family",
+            "architecture",
+            "base_addr",
+            "binary_size",
+            "binweight",
+            "bitness",
+            "component",
+            "is_library",
+            "sha256",
+            "timestamp",
+            "version",
+            "statistics.num_functions"
+        )
+        
+        sort_data = self._get_sort_data("sample_id", sort_by, is_ascending)
+        result = self._getSearchResultTemplate(
+            self._storage.findSampleByString,
+            search_term,
+            sort_data,
+            cursor,
+            limit,
+        )
+        result["id_match"] = id_match
+        result["sha_match"] = sha_match
+        return result
+
 
     ##### CONFIG CHANGES ####
     def updateMinHashThreshold(self, threshold):
