@@ -1,10 +1,17 @@
+import functools
+import operator
+import re
+import uuid
 import logging
-from collections import defaultdict
 from copy import deepcopy
-from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union
+from collections import defaultdict
+from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set, Tuple, Union
 
 from picblocks.blockhasher import BlockHasher
+from mcrit.index.SearchCursor import FullSearchCursor
 
+from mcrit.index.SearchQueryTree import AndNode, BaseVisitor, FilterSingleElementLists, NodeType, OrNode, PropagateNot, SearchConditionNode, SearchFieldResolver
+from mcrit.storage.FamilyEntry import FamilyEntry
 from mcrit.storage.FunctionEntry import FunctionEntry
 from mcrit.storage.SampleEntry import SampleEntry
 from mcrit.storage.StorageInterface import StorageInterface
@@ -19,11 +26,92 @@ if TYPE_CHECKING: # pragma: no cover
 LOGGER = logging.getLogger(__name__)
 
 
+# Helper Methods for search
+def _get_field_once(object, field):
+    try:
+        return getattr(object, field)
+    except Exception:
+        pass
+    try:
+        return object[field]
+    except Exception:
+        pass
+
+def _get_field(object, field):
+    dot_index = field.find(".")
+    if dot_index >= 0:
+        first_field = field[:dot_index]
+        remaining_fields = field[dot_index+1:]
+        return _get_field(
+            _get_field_once(object, first_field),
+            remaining_fields
+        )
+    else:
+        return _get_field_once(object, field)
+
+
+class MemorySearchTranspiler(BaseVisitor):
+    """
+    Converts a tree to a filter function that can be applied to an entry, returning True or False.
+    The input tree MUST NOT contain Not or SearchTerm nodes.
+    """
+    def visitAndNode(self, node:AndNode) -> Callable:
+        visited_children: List[Callable] = [self.visit(child) for child in node.children]
+        def and_function(entry):
+            for function in visited_children:
+                if not function(entry):
+                    return False
+            return True
+        return and_function
+
+    def visitOrNode(self, node:OrNode) -> Callable:
+        visited_children: List[Callable] = [self.visit(child) for child in node.children]
+        def or_function(entry):
+            for function in visited_children:
+                if function(entry):
+                    return True
+            return False
+        return or_function
+
+    def visitSearchConditionNode(self, node:SearchConditionNode) -> Callable:
+        string_to_operator = {
+            "<": operator.lt,
+            "<=": operator.le,
+            ">": operator.gt,
+            ">=": operator.ge,
+            "=": operator.eq,
+            "": operator.eq,
+            "!=": operator.ne,
+            "?": None,
+            "!?": None,
+        }
+        value = node.value
+        if node.operator.endswith("?"):
+            regex = re.compile(re.escape(node.value), re.IGNORECASE)
+            inverse = node.operator == "!?"
+            def check_regex(entry):
+                value = _get_field(entry, node.field)
+                if not isinstance(value, str):
+                    return False
+                return (regex.search(value) is not None) ^ inverse
+            return check_regex
+
+        if node.field in ("pichash", "offset") or node.field.endswith("_id") or "num_" in node.field:
+            try:
+                value = int(value, 0)
+            except Exception:
+                pass
+
+        chosen_operator = string_to_operator[node.operator]
+        def compare(entry):
+            return chosen_operator(_get_field(entry, node.field), value)
+        return compare
+
 class MemoryStorage(StorageInterface):
-    _families: Dict[int, str]
+    _families: Dict[int, FamilyEntry]
     _samples: Dict[int, "SampleEntry"]
     _functions: Dict[int, "FunctionEntry"]
-    _pichashes: Dict[int, Set[Tuple[int, int]]]
+    _pichashes: Dict[int, Set[Tuple[int, int, int]]]
     # Dict[band_number, Dict[?, List[function_id]]]
     _bands: Dict[int, Dict[int, List[int]]]
     _counters: Dict[str, int]
@@ -39,6 +127,8 @@ class MemoryStorage(StorageInterface):
         self._setupEmptyStorage()
 
     def _setupEmptyStorage(self) -> None:
+        self._mcrit_db_id = str(uuid.uuid4())
+        self._db_state = 0
         self._families = {}
         self._samples = {}
         self._functions = {}
@@ -51,11 +141,22 @@ class MemoryStorage(StorageInterface):
         unknown_family_id = self.addFamily("")
         assert unknown_family_id == 0
 
+    def _updateDbState(self):
+        self._db_state += 1
+
     def _useCounter(self, name: str) -> int:
         result = self._counters[name]
         self._counters[name] += 1
         return result
 
+
+    def _updateFamilyStats(self, family_id, num_samples_inc, num_functions_inc, num_library_samples_inc):
+        family_entry = self.getFamily(family_id)
+        assert family_entry is not None
+        family_entry.num_samples += num_samples_inc
+        family_entry.num_functions += num_functions_inc
+        family_entry.num_library_samples += num_library_samples_inc
+
     # TODO check if this works
     def deleteSample(self, sample_id: int) -> bool:
         if not self.isSampleId(sample_id):
@@ -67,7 +168,7 @@ class MemoryStorage(StorageInterface):
             del self._functions[function_id]
             # remove pichash entries
             if function_entry.pichash:
-                self._pichashes[function_entry.pichash].remove((sample_id, function_id))
+                self._pichashes[function_entry.pichash].remove((function_entry.family_id, sample_id, function_id))
             minhash = function_entry.getMinHash()
             # remove minhash entries, if necessary
             if not minhash:
@@ -85,15 +186,41 @@ class MemoryStorage(StorageInterface):
         del self._sample_id_to_function_ids[sample_id]
         # in case to samples with same sha256 are added and then removed, it could crash, so remove entry safely
         self._sample_by_sha256.pop(self._samples[sample_id].sha256, None)
+
+        sample_entry = self.getSampleById(sample_id)
+        self._updateFamilyStats(sample_entry.family_id, -1, -sample_entry.statistics["num_functions"], -int(sample_entry.is_library))
         # remove sample
         del self._samples[sample_id]
         return True
 
+
+    def deleteFamily(self, family_id: int, keep_samples: Optional[str] = False) -> bool:
+        if family_id not in self.families:
+            return False
+        sample_entries = self.getSamplesByFamilyId(family_id)
+        self._families.pop(family_id)
+        if keep_samples:
+            for sample_entry in sample_entries:
+                self._samples[sample_entry.sample_id]["family_id"] = 0
+                self._samples[sample_entry.sample_id]["family"] = ""
+            function_ids_to_modify = set()
+            for function_id, function_entry in self._functions.items():
+                if function_entry.family_id == family_id:
+                    function_ids_to_modify.add(function_id)
+            for function_id in function_ids_to_modify:
+                    self._functions[function_id]["family_id"] = 0
+        else:
+            for sample_entry in sample_entries:
+                self.deleteSample(sample_entry.sample_id)
+        self._updateDbState()
+        return True
+
     def addSmdaReport(self, smda_report: "SmdaReport") -> Optional["SampleEntry"]:
         sample_entry = None
         if not self.getSampleBySha256(smda_report.sha256):
+            family_id = self.addFamily(smda_report.family)
             sample_entry = SampleEntry(
-                smda_report, sample_id=self._useCounter("samples"), family_id=self.addFamily(smda_report.family)
+                smda_report, sample_id=self._useCounter("samples"), family_id=family_id
             )
             self._samples[sample_entry.sample_id] = sample_entry
             self._sample_by_sha256[sample_entry.sha256] = sample_entry.sample_id
@@ -102,6 +229,7 @@ class MemoryStorage(StorageInterface):
                 function_entry = self._addFunction(sample_entry, smda_function)
                 function_ids.append(function_entry.function_id)
             self._sample_id_to_function_ids[sample_entry.sample_id] = function_ids
+            self._updateFamilyStats(family_id, +1, sample_entry.statistics["num_functions"], int(sample_entry.is_library))
         else:
             LOGGER.warn("Sample %s already existed, skipping.", smda_report.sha256)
         return sample_entry
@@ -112,6 +240,7 @@ class MemoryStorage(StorageInterface):
             sample_entry.sample_id = sample_id
             self._samples[sample_id] = sample_entry
             self._sample_by_sha256[sample_entry.sha256] = sample_id
+            self._updateFamilyStats(sample_entry.family_id, +1, sample_entry.statistics["num_functions"], int(sample_entry.is_library))
         else:
             LOGGER.warn("Sample %s already existed, skipping.", sample_entry.sha256)
         return sample_entry
@@ -126,9 +255,24 @@ class MemoryStorage(StorageInterface):
         if function_entry.pichash:
             if function_entry.pichash not in self._pichashes:
                 self._pichashes[function_entry.pichash] = set()
-            self._pichashes[function_entry.pichash].add((function_entry.sample_id, function_entry.function_id))
+            self._pichashes[function_entry.pichash].add((function_entry.family_id, function_entry.sample_id, function_entry.function_id))
         return function_entry
 
+
+    def importFunctionEntries(self, function_entries: List["FunctionEntry"]) -> Optional[List["FunctionEntry"]]:
+        for function_entry in function_entries:
+            function_entry.function_id = self._useCounter("functions")
+            # add function to regular storage
+            self._functions[function_entry.function_id] = function_entry
+            if function_entry.sample_id not in self._sample_id_to_function_ids:
+                self._sample_id_to_function_ids[function_entry.sample_id] = []
+            self._sample_id_to_function_ids[function_entry.sample_id].append(function_entry.function_id)
+            if function_entry.pichash:
+                if function_entry.pichash not in self._pichashes:
+                    self._pichashes[function_entry.pichash] = set()
+                self._pichashes[function_entry.pichash].add((function_entry.family_id, function_entry.sample_id, function_entry.function_id))
+        return function_entries
+
     # TODO return type?
     def getLibraryInfoForSampleId(self, sample_id: int) -> Optional[Dict[str, str]]:
         if sample_id not in self._samples:
@@ -139,10 +283,25 @@ class MemoryStorage(StorageInterface):
     def _addFunction(
         self, sample_entry: "SampleEntry", smda_function: "SmdaFunction", minhash: Optional["MinHash"] = None
     ) -> "FunctionEntry":
+        """Add a function (and optionally its MinHash) to storage, using the respective SampleEntry for reference.
+
+        Args:
+            sample_entry: The SampleEntry of smda_function
+            smda_function: The SmdaFunction to be added.
+            minhash: (Optional, defaults to None) The MinHash of the Function
+
+        Returns:
+            A FunctionEntry
+        """
         function_entry = FunctionEntry(sample_entry, smda_function, self._useCounter("functions"), minhash=minhash)
         image_lower = sample_entry.base_addr
         image_upper = image_lower + sample_entry.binary_size
-        function_entry.picblockhashes = self.blockhasher.getBlockhashesForFunction(smda_function, image_lower, image_upper, hash_size=8)
+        picblockhashes = []
+        for hash_entry in self.blockhasher.getBlockhashesForFunction(smda_function, image_lower, image_upper, hash_size=8):
+            for block_entry in hash_entry["offset_tuples"]:
+                block_entry["hash"] = hash_entry["hash"]
+                picblockhashes.append(block_entry)
+        function_entry.picblockhashes = picblockhashes
         self._functions[function_entry.function_id] = function_entry
         if minhash and minhash.hasMinHash():
             minhash.function_id = function_entry.function_id
@@ -150,7 +309,7 @@ class MemoryStorage(StorageInterface):
         if function_entry.pichash:
             if function_entry.pichash not in self._pichashes:
                 self._pichashes[function_entry.pichash] = set()
-            self._pichashes[function_entry.pichash].add((sample_entry.sample_id, function_entry.function_id))
+            self._pichashes[function_entry.pichash].add((function_entry.family_id, sample_entry.sample_id, function_entry.function_id))
         return function_entry
 
     def addMinHash(self, minhash: "MinHash") -> bool:
@@ -200,19 +359,28 @@ class MemoryStorage(StorageInterface):
     def isPicHash(self, pichash: int) -> bool:
         return pichash in self._pichashes
 
-    def getMatchesForPicHash(self, pichash: int) -> Optional[Set[Tuple[int, int]]]:
+    def getMatchesForPicHash(self, pichash: int) -> Optional[Set[Tuple[int, int, int]]]:
         if pichash not in self._pichashes:
             return set()
         return deepcopy(self._pichashes[pichash])
 
-    def getFamily(self, family_id: int) -> Optional[str]:
+
+    def getMatchesForPicBlockHash(self, picblockhash: int) -> Optional[Set[Tuple[int, int, int, int]]]:
+        result = set()
+        for function_id, function_entry in self._functions.items():
+            for pbh in function_entry.picblockhashes:
+                if picblockhash == pbh["hash"]:
+                    result.add([function_entry.family_id, function_entry.sample_id, function_id, pbh["offset"]])
+        return result
+
+    def getFamily(self, family_id: int) -> Optional[FamilyEntry]:
         if family_id in self._families:
-            return deepcopy(self._families[family_id])
+            return self._families[family_id]
         return None
 
     def getFamilyId(self, family_name: str) -> Optional[int]:
-        for fam_id, fam_name in self._families.items():
-            if fam_name == family_name:
+        for fam_id, fam_entry in self._families.items():
+            if fam_entry.family_name == family_name:
                 return fam_id
         return None
 
@@ -220,12 +388,15 @@ class MemoryStorage(StorageInterface):
         family_id = self.getFamilyId(family_name)
         if family_id is None:
             family_id = self._useCounter("families")
-            self._families[family_id] = family_name
+            self._families[family_id] = FamilyEntry(family_name = family_name, family_id=family_id)
         return family_id
 
-    def getFunctionById(self, function_id: int) -> Optional["FunctionEntry"]:
+    def getFunctionById(self, function_id: int, with_xcfg=False) -> Optional["FunctionEntry"]:
         if function_id in self._functions:
-            return deepcopy(self._functions[function_id])
+            function_entry = deepcopy(self._functions[function_id])
+            if with_xcfg is False:
+                function_entry.xcfg = None
+            return deepcopy(function_entry)
         return None
 
     # TODO does this need to be more efficient?
@@ -354,7 +525,7 @@ class MemoryStorage(StorageInterface):
     # TODO dict of what?
     def getContent(self) -> Dict[str, Any]:
         content = {
-            "families": self._families,
+            "families": {family_id: family.toDict() for family_id, family in self._families.items()},
             "samples": {sample_id: sample.toDict() for sample_id, sample in self._samples.items()},
             "functions": {function_id: function.toDict() for function_id, function in self._functions.items()},
             "bands": self._bands,
@@ -363,7 +534,7 @@ class MemoryStorage(StorageInterface):
 
     # TODO dict of what?
     def setContent(self, content: Dict[str, Any]) -> None:
-        self._families = {int(k): v for k, v in content["families"].items()}
+        self._families = {int(k): FamilyEntry.fromDict(v) for k, v in content["families"].items()}
         self._samples = {int(k): SampleEntry.fromDict(v) for k, v in content["samples"].items()}
         self._bands = {int(k): {int(ik): iv for ik, iv in v.items()} for k, v in content["bands"].items()}
         self._sample_by_sha256 = {sample.sha256: sample_id for sample_id, sample in self._samples.items()}
@@ -381,6 +552,7 @@ class MemoryStorage(StorageInterface):
 
     def getStats(self) -> Dict[str, Union[int, Dict[int, int]]]:
         stats = {
+            "db_state": self._db_state,
             "num_families": len(self._families),
             "num_samples": len(self._samples),
             "num_functions": len(self._functions),
@@ -395,38 +567,92 @@ class MemoryStorage(StorageInterface):
             if band_hash not in self._bands[band_number]:
                 self._bands[band_number][band_hash] = []
             self._bands[band_number][band_hash].append(minhash.function_id)
+            
+    ##### helpers for search ######
 
-    def findFamilyByString(self, needle: str, max_num_results: int = 100) -> Dict[int, str]:
+    @staticmethod
+    def _get_sort_key_from_cursor(full_cursor: Optional[FullSearchCursor]):
+        assert full_cursor is not None
+
+        # Mapping entry -> tuple to sort by
+        def get_tuple_from_entry(entry):
+            return tuple(_get_field(entry, field) for field in full_cursor.sort_fields)
+
+        # custom tuple compare function considering sort direction per dimension 
+        is_backward_search = not full_cursor.is_forward_search
+        sort_directions = [1 if direction ^ is_backward_search else -1 for direction in full_cursor.sort_directions]
+        def compare_tuple(tuple1, tuple2):
+            for i in range(len(full_cursor.sort_fields)):
+                direction = sort_directions[i]
+                val1 = tuple1[i]
+                val2 = tuple2[i]
+                if val1 == val2:
+                    pass
+                if val1 is None:
+                    return -direction
+                if val2 is None:
+                    return +direction
+                if val1 < val2:
+                    return -direction
+                if val1 > val2:
+                    return direction
+            return 0
+
+        return lambda entry: functools.cmp_to_key(compare_tuple)(get_tuple_from_entry(entry))
+
+
+    def _get_search_filter(self, search_fields:List[str], search_tree: NodeType, cursor: Optional[FullSearchCursor], conditional_search_fields=None) -> Callable:
+        if cursor is not None:
+            full_tree = AndNode([search_tree, cursor.toTree()])
+        else:
+            full_tree = search_tree
+        full_tree = SearchFieldResolver(search_fields, conditional_search_fields=conditional_search_fields).visit(full_tree)
+        full_tree = FilterSingleElementLists().visit(full_tree)
+        full_tree = PropagateNot().visit(full_tree)
+        filter = MemorySearchTranspiler().visit(full_tree)
+        return filter
+
+
+    ##### search ######
+
+    # NOTE: Slowest part of search (by far) is the sorting. Maybe this can be precomputed or cached?
+    # NOTE: Once sorting is fast, the search cursor could be handled separately from the search tree.
+    #       Instead of putting it in the filter function, we could apply binary search on the sorted data.
+
+    def findFamilyByString(self, search_tree: NodeType, cursor: Optional[FullSearchCursor] = None, max_num_results: int = 100) -> Dict[int, "FamilyEntry"]:
         result_dict = {}
-        self._families
-        for family_id, family_name in self._families.items():
-            if needle in family_name:
-                result_dict[family_id] = family_name
-            if len(result_dict) > max_num_results:
+        search_fields = ["family_name"]
+        filter = self._get_search_filter(search_fields, search_tree, cursor)
+        sort_key = self._get_sort_key_from_cursor(cursor)
+        for entry in sorted(self._families.values(), key=sort_key):
+            if filter(entry):
+                result_dict[entry.family_id] = entry
+            if len(result_dict) >= max_num_results:
                 break
         return result_dict
 
-    def findSampleByString(self, needle: str, max_num_results: int = 100) -> Dict[int, "SampleEntry"]:
+    def findSampleByString(self, search_tree: NodeType, cursor: Optional[FullSearchCursor] = None, max_num_results: int = 100) -> Dict[int, "SampleEntry"]:
         result_dict = {}
-        for sample_id, sample_entry in self._samples.items():
-            if needle in sample_entry.filename:
-                result_dict[sample_id] = sample_entry
-            elif len(needle) >= 3 and needle in entry.sha256:
+        search_fields = ["filename", "family", "component", "version",]
+        conditional_field = ("sha256", lambda search_term: len(search_term)>=3)
+        filter = self._get_search_filter(search_fields, search_tree, cursor, conditional_search_fields=[conditional_field])
+        sort_key = self._get_sort_key_from_cursor(cursor)
+        for entry in sorted(self._samples.values(), key=sort_key):
+            if filter(entry):
                 result_dict[entry.sample_id] = entry
-            elif needle in sample_entry.component:
-                result_dict[sample_id] = sample_entry
-            elif needle in sample_entry.version:
-                result_dict[sample_id] = sample_entry
-            if len(result_dict) > max_num_results:
+            if len(result_dict) >= max_num_results:
                 break
         return result_dict
 
-    def findFunctionByString(self, needle: str, max_num_results: int = 100) -> Dict[int, "FunctionEntry"]:
+    def findFunctionByString(self, search_tree: NodeType, cursor: Optional[FullSearchCursor] = None, max_num_results: int = 100) -> Dict[int, "FunctionEntry"]:
         result_dict = {}
-        for function_id, function_entry in self._functions.items():
-            if needle in function_entry.function_name:
-                result_dict[function_id] = function_entry
         # TODO also search through function labels once we have implemented them
-            if len(result_dict) > max_num_results:
+        search_fields = ["function_name"]
+        filter = self._get_search_filter(search_fields, search_tree, cursor)
+        sort_key = self._get_sort_key_from_cursor(cursor)
+        for entry in sorted(self._functions.values(), key=sort_key):
+            if filter(entry):
+                result_dict[entry.function_id] = entry
+            if len(result_dict) >= max_num_results:
                 break
-        return result_dict
+        return result_dict
\ No newline at end of file
