@@ -1,10 +1,11 @@
-import datetime
+import re
+import uuid
 import json
 import logging
+import datetime
 import traceback
 from operator import itemgetter
-from typing import Any, TYPE_CHECKING, Dict, List, Optional, Set, Tuple
-
+from typing import Any, TYPE_CHECKING, Dict, Iterable, List, Optional, Set, Tuple, Union
 
 LOGGER = logging.getLogger(__name__)
 try:
@@ -15,7 +16,10 @@ except:
 
 from picblocks.blockhasher import BlockHasher
 
+from mcrit.index.SearchCursor import FullSearchCursor
+from mcrit.index.SearchQueryTree import AndNode, BaseVisitor, FilterSingleElementLists, NodeType, NotNode, OrNode, PropagateNot, SearchConditionNode, SearchFieldResolver, SearchTermNode
 from mcrit.libs.utility import generate_unique_groups
+from mcrit.storage.FamilyEntry import FamilyEntry
 from mcrit.storage.FunctionEntry import FunctionEntry
 from mcrit.storage.MatchingCache import MatchingCache
 from mcrit.storage.SampleEntry import SampleEntry
@@ -31,6 +35,65 @@ if TYPE_CHECKING: # pragma: no cover
 
 # TODO when is checking self._database.samples.count_documents() necessary?
 
+class MongoSearchTranspiler(BaseVisitor):
+    """
+    Converts a tree to a MongoDB query.
+    The input tree MUST NOT contain Not or SearchTerm nodes.
+    """
+    @staticmethod
+    def _or_query(*conditions):
+        if len(conditions) == 0:
+            return {}
+        if len(conditions) == 1:
+            return conditions[0]
+        return {"$or": conditions}
+
+    @staticmethod
+    def _and_query(*conditions):
+        if len(conditions) == 0:
+            return {}
+        if len(conditions) == 1:
+            return conditions[0]
+        return {"$and": conditions}
+
+    def visitAndNode(self, node: AndNode):
+        visited_children = [self.visit(child) for child in node.children]
+        return self._and_query(*visited_children)
+
+    def visitOrNode(self, node: OrNode):
+        visited_children = [self.visit(child) for child in node.children]
+        return self._or_query(*visited_children)
+
+    def visitSearchConditionNode(self, node:SearchConditionNode):
+        operator_to_mongo = {
+            "<": "$lt",
+            "<=": "$lte",
+            ">": "$gt",
+            ">=": "$gte",
+            "=": None,
+            "": None,
+            "!=": "$ne",
+            "?": None,
+            "!?": "$not",
+        }
+        value = node.value
+        if node.operator.endswith("?"):
+            value = re.compile(re.escape(node.value), re.IGNORECASE)
+        elif node.field in ("pichash", "offset") or node.field.endswith("_id") or "num_" in node.field:
+            try:
+                value = int(value, 0)
+            except Exception:
+                pass
+        mongo_operator = operator_to_mongo[node.operator]
+        if mongo_operator is None:
+            condition = {node.field: value}
+        else:
+            condition = {node.field: {mongo_operator: value}}
+        # TODO: fix
+        MongoDbStorage._encodePichash(None, condition)
+        return condition
+
+
 
 class MongoDbStorage(StorageInterface):
 
@@ -50,12 +113,15 @@ class MongoDbStorage(StorageInterface):
         self._ensureIndexAndUnknownFamily()
 
     def _ensureIndexAndUnknownFamily(self) -> None:
+        if "settings" not in self._database.list_collection_names():
+            self._database["settings"].insert_one({"mcrit_db_id": str(uuid.uuid4()), "db_state": 0})
         self._database["samples"].create_index("sample_id")
         self._database["families"].create_index("family_id")
         self._database["functions"].create_index("function_id")
         self._database["functions"].create_index("sample_id")
         self._database["functions"].create_index("_pichash")
         self._database["functions"].create_index("_picblockhashes.hash")
+        self._database["functions"].create_index("_picblockhashes.offset")
         self._database["matches"].create_index("match_id")
         self._database["candidates"].create_index("function_id")
         self._database["counters"].create_index("name")
@@ -64,7 +130,7 @@ class MongoDbStorage(StorageInterface):
         # Add Family "" if it is not already in storage
         if self.getFamily(0) is None:
             self.addFamily("")
-        assert self.getFamily(0) == ""
+        assert self.getFamily(0).family_name == ""
 
     ###############################################################################
     # Generic database functionality and logging
@@ -111,6 +177,20 @@ class MongoDbStorage(StorageInterface):
             )
             raise ValueError("Database insert failed.")
 
+
+    def _dbInsertMany(self, collection: str, data: List["Dict"]):
+        try:
+            insert_result = self._database[collection].insert_many([self._toBinary(document) for document in data])
+            if insert_result.acknowledged:
+                return insert_result.inserted_ids
+            return None
+        except Exception as exc:
+            self._dbLogError(
+                'Database insert_many for collection "%s" failed.' % collection,
+                {"user": "internal/mcrit", "traceback": traceback.format_exc(exc).split("\n")},
+            )
+            raise ValueError("Database insert failed.")
+
     def _dbQuery(self, collection: str, query: Dict, find_one: bool = False):
         try:
             if find_one:
@@ -138,16 +218,36 @@ class MongoDbStorage(StorageInterface):
         record = {"ts": self._getCurrentTimestamp(), "%s_msg" % log_type: log_msg, "%s_details" % log_type: log_details}
         self._database[log_type].insert_one(self._toBinary(record))
 
-    def _useCounter(self, name: str) -> int:
-        result = self._database.counters.find_one_and_update(
+    def _useCounterBulk(self, name: str, num_counts: int) -> Iterable[int]:
+        assert num_counts >= 0
+        if num_counts == 0:
+            return []
+        query_result = self._database.counters.find_one_and_update(
             filter={"name": name}, 
-            update={"$inc": {"value": 1}}, 
-            # return_document=ReturnDocument.AFTER,
+            update={"$inc": {"value": num_counts}}, 
             upsert=True
         )
+        first_count = 0
+        if query_result is not None:
+            first_count = query_result["value"]
+        return range(first_count, first_count+num_counts)
+    
+    def _useCounter(self, name:str) -> int:
+        return next(iter(self._useCounterBulk(name, 1)))
+
+    def _updateDbState(self):
+        result = self._database.settings.find_one_and_update({}, { "$inc": { "db_state": 1}})
         if result is None:
-            return 0
-        return result["value"]
+            raise Exception("Database does not have a db_state field")
+        else:
+            return result["db_state"]
+
+    def _getDbState(self):
+        result = self._database.settings.find_one({})
+        if result is None:
+            raise Exception("Database does not have a db_state field")
+        else:
+            return result["db_state"]
 
     ###############################################################################
     # Conversion
@@ -171,12 +271,10 @@ class MongoDbStorage(StorageInterface):
                 del function_dict["pichash"]
         if "picblockhashes" in function_dict:
             converted_entries = []
-            for blockhash_dict in function_dict["picblockhashes"]:
-                converted_entries.append({
-                    "hash": hex(blockhash_dict["hash"]),
-                    "size": blockhash_dict["size"],
-                    "count": blockhash_dict["count"],
-                })
+            for entry in function_dict["picblockhashes"]:
+                converted_entry = dict(**entry)
+                converted_entry["hash"] = hex(converted_entry["hash"])
+                converted_entries.append(converted_entry)
             function_dict["_picblockhashes"] = converted_entries
             if delete_old:
                 del function_dict["picblockhashes"]
@@ -188,12 +286,10 @@ class MongoDbStorage(StorageInterface):
                 del function_dict["_pichash"]
         if "_picblockhashes" in function_dict:
             converted_entries = []
-            for blockhash_dict in function_dict["_picblockhashes"]:
-                converted_entries.append({
-                    "hash": int(blockhash_dict["hash"], 16),
-                    "size": blockhash_dict["size"],
-                    "count": blockhash_dict["count"],
-                })
+            for entry in function_dict["_picblockhashes"]:
+                converted_entry = dict(**entry)
+                converted_entry["hash"] = int(converted_entry["hash"], 16)
+                converted_entries.append(converted_entry)
             function_dict["picblockhashes"] = converted_entries
             if delete_old:
                 del function_dict["_picblockhashes"]
@@ -231,6 +327,7 @@ class MongoDbStorage(StorageInterface):
         return function_document["sample_id"]
 
     def deleteSample(self, sample_id: int) -> bool:
+        # TODO the way of removing function_ids from band hashes here is super inefficient and slow
         function_entries = self.getFunctionsBySampleId(sample_id)
         if function_entries is None:
             # in this case sample_id is does not exist
@@ -239,7 +336,7 @@ class MongoDbStorage(StorageInterface):
         for function_entry in function_entries:
             minhash = function_entry.getMinHash()
             # remove minhash entries, if necessary
-            if not minhash:
+            if not minhash or not minhash.hasMinHash():
                 continue
             band_hashes = self.getBandHashesForMinHash(minhash)
             for band_number, band_hash in sorted(band_hashes.items()):
@@ -253,10 +350,42 @@ class MongoDbStorage(StorageInterface):
                 if band_document is not None and len(band_document["function_ids"]) == 0:
                     self._database["band_%d" % band_number].delete_one({"band_hash": band_hash})
 
+        # update family stats
+        sample_entry = self.getSampleById(sample_id)
+        self._updateFamilyStats(sample_entry.family_id, -1, -sample_entry.statistics["num_functions"], -int(sample_entry.is_library))
         # remove functions
         self._database.functions.delete_many({"sample_id": sample_id})
         # remove sample
         self._database.samples.delete_one({"sample_id": sample_id})
+        self._updateDbState()
+        return True
+
+    def _updateFamilyStats(self, family_id, num_samples_inc, num_functions_inc, num_library_samples_inc):
+        self._database.families.update_one(
+            {"family_id": family_id},
+            {
+                "$inc": {
+                    "num_samples": num_samples_inc,
+                    "num_functions": num_functions_inc,
+                    "num_library_samples": num_library_samples_inc,
+                },
+            }
+        )
+
+    def deleteFamily(self, family_id: int, keep_samples: Optional[str] = False) -> bool:
+        family_document = self._database.families.find_one({"family_id": family_id})
+        family_name = family_document["family_name"]
+        if family_document is None:
+            return False
+        sample_entries = self.getSamplesByFamilyId(family_id)
+        self._database.families.delete_one({"family_id": family_id})
+        if keep_samples:
+            self._database.samples.update_many({"family_id": family_id}, {"$set": {"family_id": 0, "family": ""}})
+            self._database.functions.update_many({"family_id": family_id}, {"$set": {"family_id": 0}})
+        else:
+            for sample_entry in sample_entries:
+                self.deleteSample(sample_entry.sample_id)
+        self._updateDbState()
         return True
 
     def getSamplesByFamilyId(self, family_id: int) -> Optional[List["SampleEntry"]]:
@@ -290,12 +419,18 @@ class MongoDbStorage(StorageInterface):
     def addSmdaReport(self, smda_report: "SmdaReport") -> Optional["SampleEntry"]:
         sample_entry = None
         if not self.getSampleBySha256(smda_report.sha256):
+            family_id = self.addFamily(smda_report.family)
             sample_entry = SampleEntry(
-                smda_report, sample_id=self._useCounter("samples"), family_id=self.addFamily(smda_report.family)
+                smda_report, sample_id=self._useCounter("samples"), family_id=family_id
             )
             self._dbInsert("samples", sample_entry.toDict())
-            for smda_function in smda_report.getFunctions():
-                self._addFunction(sample_entry, smda_function)
+            function_ids = self._useCounterBulk("functions",  smda_report.num_functions)
+            function_dicts = []
+            for function_id, smda_function in zip(function_ids, smda_report.getFunctions()):
+                function_dicts.append(self._getFunctionDocument(sample_entry, smda_function, function_id))
+            self._dbInsertMany("functions", function_dicts)
+            self._updateFamilyStats(family_id, +1, sample_entry.statistics["num_functions"], int(sample_entry.is_library))
+            self._updateDbState()
         else:
             LOGGER.warn("Sample %s already existed, skipping.", smda_report.sha256)
         return sample_entry
@@ -305,6 +440,8 @@ class MongoDbStorage(StorageInterface):
             sample_id = self._useCounter("samples")
             sample_entry.sample_id = sample_id
             self._dbInsert("samples", sample_entry.toDict())
+            self._updateFamilyStats(sample_entry.family_id, +1, sample_entry.statistics["num_functions"], int(sample_entry.is_library))
+            self._updateDbState()
         else:
             LOGGER.warn("Sample %s already existed, skipping.", sample_entry.sha256)
         return sample_entry
@@ -317,6 +454,18 @@ class MongoDbStorage(StorageInterface):
         self._dbInsert("functions", function_dict)
         return function_entry
 
+    def importFunctionEntries(self, function_entries: List["FunctionEntry"]) -> Optional[List["FunctionEntry"]]:
+        functions_as_dicts = []
+        function_ids = self._useCounterBulk("functions", len(function_entries))
+        for function_id, function_entry in zip(function_ids, function_entries):
+            function_entry.function_id = function_id
+            # add function to regular storage
+            function_dict = function_entry.toDict()
+            self._encodeFunction(function_dict)
+            functions_as_dicts.append(function_dict)
+        self._dbInsertMany("functions", functions_as_dicts)
+        return function_entries
+
     def getFunctionsBySampleId(self, sample_id: int) -> Optional[List["FunctionEntry"]]:
         if not self.isSampleId(sample_id):
             return None
@@ -351,7 +500,7 @@ class MongoDbStorage(StorageInterface):
         )  # ["function_id"]
         return self.getPicHashMatchesByFunctionIds(function_ids)
 
-    def getPicHashMatchesByFunctionId(self, function_id: int) -> Optional[Dict[int, Set[Tuple[int, int]]]]:
+    def getPicHashMatchesByFunctionId(self, function_id: int) -> Optional[Dict[int, Set[Tuple[int, int, int]]]]:
         query_result = self._database.functions.find_one({"function_id": function_id})
         if query_result is None or not "_pichash" in query_result:
             return None
@@ -363,14 +512,14 @@ class MongoDbStorage(StorageInterface):
 
         sample_and_function_ids = set(
             map(
-                lambda x: (x["sample_id"], x["function_id"]),
+                lambda x: (x["family_id"], x["sample_id"], x["function_id"]),
                 list(self._database.functions.find({"_pichash": encoded_pichash})),
             )
         )
 
         return {decoded_pichash: sample_and_function_ids}
 
-    def getPicHashMatchesByFunctionIds(self, function_ids: List[int]) -> Dict[int, Set[Tuple[int, int]]]:
+    def getPicHashMatchesByFunctionIds(self, function_ids: List[int]) -> Dict[int, Set[Tuple[int, int, int]]]:
         # internal format!
         pichash_raw = self._database.functions.find({"function_id": {"$in": function_ids}}, {"_pichash": 1, "_id": 0})
         pichashes_list = list(pichash_raw)  # broken? TODO
@@ -384,7 +533,7 @@ class MongoDbStorage(StorageInterface):
                 continue
             pichashes[decoded_pichash] = set(
                 map(
-                    lambda x: (x["sample_id"], x["function_id"]),
+                    lambda x: (x["family_id"], x["sample_id"], x["function_id"]),
                     list(self._database.functions.find({"_pichash": encoded_pichash})),
                 )
             )
@@ -435,13 +584,13 @@ class MongoDbStorage(StorageInterface):
     def _addMinHashesToBands(self, minhashes: List["MinHash"]) -> None:
         band_hashes = {number: {} for number in range(self._config.STORAGE_NUM_BANDS)}
         for minhash in minhashes:
-            for band_number, band_hash in self.getBandHashesForMinHash(minhash).items():
-                if band_hash not in band_hashes[band_number]:
-                    band_hashes[band_number][band_hash] = []
-                band_hashes[band_number][band_hash].append(minhash.function_id)
+            if minhash.hasMinHash():
+                for band_number, band_hash in self.getBandHashesForMinHash(minhash).items():
+                    if band_hash not in band_hashes[band_number]:
+                        band_hashes[band_number][band_hash] = []
+                    band_hashes[band_number][band_hash].append(minhash.function_id)
         self._updateBands(band_hashes)
 
-    # original implementation, already working
     def _updateBands(self, band_hashes: Dict[int, Dict[int, List[int]]]) -> None:
         num_band_updates = 0
         for band_number, band_data in band_hashes.items():
@@ -453,15 +602,33 @@ class MongoDbStorage(StorageInterface):
             self._database["band_%d" % band_number].bulk_write(band_updates, ordered=False)
             num_band_updates += len(band_updates)
 
-    # COPIED FROM MEMORYSTORAGE
-    # TODO optimize or move to interface
     def getCandidatesForMinHashes(self, function_id_to_minhash: Dict[int, "MinHash"]) -> Dict[int, Set[int]]:
         candidates = {}
+        target_band_hashes_per_band = {band_number: set() for band_number in range(self._config.STORAGE_NUM_BANDS)}
+        band_hash_to_function_ids = {band_number: {} for band_number in range(self._config.STORAGE_NUM_BANDS)}
         for function_id, minhash in function_id_to_minhash.items():
-            candidates[function_id] = self.getCandidatesForMinHash(minhash)
+            if not minhash.hasMinHash():
+                continue
+            band_hashes = self.getBandHashesForMinHash(minhash)
+            for band_number, band_hash in sorted(band_hashes.items()):
+                target_band_hashes_per_band[band_number].add(band_hash)
+                if band_hash not in band_hash_to_function_ids[band_number]:
+                    band_hash_to_function_ids[band_number][band_hash] = set()
+                band_hash_to_function_ids[band_number][band_hash].add(function_id)
+        for band_number, band_hashes in target_band_hashes_per_band.items():
+            match_query = {"$match": {"band_hash": {"$in": list(band_hashes)}}}
+            cursor = self._database["band_%d" % band_number].aggregate([match_query])
+            for hit in cursor:
+                reference_function_ids = band_hash_to_function_ids[band_number][hit["band_hash"]]
+                for function_id in reference_function_ids:
+                    if function_id not in candidates:
+                        candidates[function_id] = set()
+                    candidates[function_id].update(hit["function_ids"])
         return candidates
 
     def getCandidatesForMinHash(self, minhash: "MinHash") -> Set[int]:
+        if not minhash.hasMinHash():
+            return
         candidates = set([])
         band_hashes = self.getBandHashesForMinHash(minhash)
         for band_number, band_hash in sorted(band_hashes.items()):
@@ -508,35 +675,53 @@ class MongoDbStorage(StorageInterface):
         self._encodePichash(query)
         return self._database.functions.find_one(query) is not None
 
-    def getMatchesForPicHash(self, pichash: int) -> Set[Tuple[int, int]]:
+    def getMatchesForPicHash(self, pichash: int) -> Set[Tuple[int, int, int]]:
         query = {"pichash": pichash}
         self._encodePichash(query)
         return set(
             map(
-                lambda x: (x["sample_id"], x["function_id"]),
-                list(self._database.functions.find(query, {"function_id": 1, "sample_id": 1, "_id": 0})),
+                lambda x: (x["family_id"], x["sample_id"], x["function_id"]),
+                list(self._database.functions.find(query, {"family_id": 1, "function_id": 1, "sample_id": 1, "_id": 0})),
+            )
+        )
+
+    def getMatchesForPicBlockHash(self, picblockhash: int) -> Set[Tuple[int, int, int, int]]:
+        query = {"_picblockhashes.hash": hex(picblockhash)}
+        result = self._database.functions.aggregate([
+            {"$match": query}, 
+            {"$unwind": "$_picblockhashes"}, 
+            {"$match": query}, 
+            {"$project": {"_id": 0, "function_id": 1, "family_id": 1, "sample_id": 1, "offset": "$_picblockhashes.offset"}}])
+        return set(
+            map(
+                lambda x: (x["family_id"], x["sample_id"], x["function_id"], x["offset"]),
+                result,
             )
         )
 
     ########################## 'old' implementations below
 
-    def _addFunction(
-        self, sample_entry: "SampleEntry", smda_function: "SmdaFunction", minhash: Optional["MinHash"] = None
-    ) -> "FunctionEntry":
-        function_id = self._useCounter("functions")
-        function_entry = FunctionEntry(sample_entry, smda_function, function_id, minhash=minhash)
+    def _getFunctionDocument(
+        self, sample_entry: "SampleEntry", smda_function: "SmdaFunction", function_id: int) -> Dict:
+        """
+        Converts a SmdaFunction to a dict ready for insertion into MongoDb.
+        Picblockhashes will be added to the function.
+        The function_id has to be explicitly provided.
+        """
+        function_entry = FunctionEntry(sample_entry, smda_function, function_id)
         # calculate block hashes and add separately
         image_lower = sample_entry.base_addr
         image_upper = image_lower + sample_entry.binary_size
-        function_entry.picblockhashes = self.blockhasher.getBlockhashesForFunction(smda_function, image_lower, image_upper, hash_size=8)
+        picblockhashes = []
+        for hash_entry in self.blockhasher.getBlockhashesForFunction(smda_function, image_lower, image_upper, hash_size=8):
+            for block_entry in hash_entry["offset_tuples"]:
+                block_entry["hash"] = hash_entry["hash"]
+                picblockhashes.append(block_entry)
+        function_entry.picblockhashes = picblockhashes
         # convert for persistance
         function_dict = function_entry.toDict()
         self._encodeFunction(function_dict)
-        self._dbInsert("functions", function_dict)
-        if minhash and minhash.hasMinHash():
-            minhash.function_id = function_entry.function_id
-            self._addMinHashToBands(minhash)
-        return function_entry
+        return function_dict
 
     def createMatchingCache(self, function_ids: List[int]) -> MatchingCache:
         cache_data = self._getCacheDataForFunctionIds(function_ids)
@@ -561,17 +746,22 @@ class MongoDbStorage(StorageInterface):
         if family_document is not None:
             return family_document["family_id"]
         family_id = self._useCounter("families")
-        self._dbInsert("families", {"family_name": family_name, "family_id": family_id})
+        family_entry = FamilyEntry(family_name=family_name, family_id=family_id)
+        self._dbInsert("families", family_entry.toDict())
+        self._updateDbState()
         return family_id
 
-    def getFamily(self, family_id: int) -> Optional[str]:
+    def getFamily(self, family_id: int) -> Optional[FamilyEntry]:
         family_document = self._database.families.find_one({"family_id": family_id})
         if family_document is None:
             return None
-        return family_document["family_name"]
+        return FamilyEntry.fromDict(family_document)
 
-    def getFunctionById(self, function_id: int) -> Optional["FunctionEntry"]:
-        function_document = self._database.functions.find_one({"function_id": function_id}, {"_id": 0})
+    def getFunctionById(self, function_id: int, with_xcfg=False) -> Optional["FunctionEntry"]:
+        field_selection = {"_id": 0}
+        if not with_xcfg:
+            field_selection["_xcfg"] = 0
+        function_document = self._database.functions.find_one({"function_id": function_id}, field_selection)
         if not function_document:
             return None
         self._decodeFunction(function_document)
@@ -594,6 +784,7 @@ class MongoDbStorage(StorageInterface):
         for result in self._database["functions"].aggregate([{"$group": {"_id": "$_pichash"}}, {"$count": "Total" }]):
             num_unique_pichashes = result["Total"]
         stats = {
+            "db_state": self._getDbState(),
             "num_families": self._database.families.count_documents(filter={}),
             "num_samples": self._database.samples.count_documents(filter={}),
             "num_functions": self._database.functions.count_documents(filter={}),
@@ -606,7 +797,7 @@ class MongoDbStorage(StorageInterface):
     def getUnhashedFunctions(self, function_ids: Optional[List[int]] = None) -> List["FunctionEntry"]:
         unhashed_functions = []
         search_query = {}
-        if function_ids is None:
+        if function_ids is not None:
             search_query = {"function_id": {"$in": list(function_ids)}}
         for function_document in self._database.functions.find(search_query, {"_id": 0}):
             self._decodeFunction(function_document)
@@ -614,41 +805,60 @@ class MongoDbStorage(StorageInterface):
                 unhashed_functions.append(FunctionEntry.fromDict(function_document))
         return unhashed_functions
 
-    def findFamilyByString(self, needle: str, max_num_results: int = 100) -> Dict[int, str]:
+
+    ##### helpers for search ######
+
+    @staticmethod
+    def _get_sort_list_from_cursor(full_cursor: Optional[FullSearchCursor]):
+        if full_cursor is None:
+            return None
+        is_backward_search = not full_cursor.is_forward_search
+        sort_list = [(key, 1 if direction ^ is_backward_search else -1) for key, direction in full_cursor.sort_by_list]
+        return sort_list
+
+
+    def _get_search_query(self, search_fields:List[str], search_tree: NodeType, cursor: Optional[FullSearchCursor], conditional_search_fields=None):
+        if cursor is not None:
+            full_tree = AndNode([search_tree, cursor.toTree()])
+        else:
+            full_tree = search_tree
+        full_tree = SearchFieldResolver(search_fields, conditional_search_fields=conditional_search_fields).visit(full_tree)
+        full_tree = FilterSingleElementLists().visit(full_tree)
+        full_tree = PropagateNot().visit(full_tree)
+        query = MongoSearchTranspiler().visit(full_tree)
+        return query
+
+    ##### search ####
+
+    def findFamilyByString(self, search_tree: NodeType, cursor: Optional[FullSearchCursor] = None, max_num_results: int = 100) -> Dict[int, "FamilyEntry"]:
         result_dict = {}
-        for family_document in self._database.families.find():
-            if needle in family_document["family_name"]:
-                result_dict[family_document["family_id"]] = family_document["family_name"]
-            if len(result_dict) > max_num_results:
-                break
+        search_fields = ["family_name"]
+        query = self._get_search_query(search_fields, search_tree, cursor)
+        sort_list = self._get_sort_list_from_cursor(cursor)
+        for family_document in self._database.families.find(query, {"_id":0}, sort=sort_list, limit=max_num_results):
+            entry = FamilyEntry.fromDict(family_document)
+            result_dict[family_document["family_id"]] = entry
         return result_dict
 
-    def findSampleByString(self, needle: str, max_num_results: int = 100) -> Dict[int, "SampleEntry"]:
+    def findSampleByString(self, search_tree: NodeType, cursor: Optional[FullSearchCursor] = None, max_num_results: int = 100) -> Dict[int, "SampleEntry"]:
         result_dict = {}
-        for sample_document in self._database.samples.find():
+        search_fields = ["filename", "family", "component", "version",]
+        conditional_field = ("sha256", lambda search_term: len(search_term)>=3)
+        query = self._get_search_query(search_fields, search_tree, cursor, conditional_search_fields=[conditional_field])
+        sort_list = self._get_sort_list_from_cursor(cursor)
+        for sample_document in self._database.samples.find(query, {"_id":0}, sort=sort_list, limit=max_num_results):
             entry = SampleEntry.fromDict(sample_document)
-            if needle in entry.filename:
-                result_dict[entry.sample_id] = entry
-            elif len(needle) >= 3 and needle in entry.sha256:
-                result_dict[entry.sample_id] = entry
-            elif needle in entry.family:
-                result_dict[entry.sample_id] = entry
-            elif needle in entry.component:
-                result_dict[entry.sample_id] = entry
-            elif needle in entry.version:
-                result_dict[entry.sample_id] = entry
-            if len(result_dict) > max_num_results:
-                break
+            result_dict[entry.sample_id] = entry
         return result_dict
 
-    def findFunctionByString(self, needle: str, max_num_results: int = 100) -> Dict[int, "FunctionEntry"]:
+    def findFunctionByString(self, search_tree: NodeType, cursor: Optional[FullSearchCursor] = None, max_num_results: int = 100) -> Dict[int, "FunctionEntry"]:
         result_dict = {}
-        for function_document in self._database.functions.find():
+        # TODO also search through function labels once we have implemented them
+        search_fields = ["function_name"]
+        query = self._get_search_query(search_fields, search_tree, cursor)
+        sort_list = self._get_sort_list_from_cursor(cursor)
+        for function_document in self._database.functions.find(query, {"_id":0, "_xcfg":0}, sort=sort_list, limit=max_num_results):
             self._decodeFunction(function_document)
             entry = FunctionEntry.fromDict(function_document)
-            if needle in entry.function_name:
-                result_dict[entry.function_id] = entry
-        # TODO also search through function labels once we have implemented them
-            if len(result_dict) > max_num_results:
-                break
+            result_dict[entry.function_id] = entry
         return result_dict
