@@ -25,6 +25,7 @@ Sha256 = str
 
 class StorageInterface:
     _config: "StorageConfig"
+    _band_projection: None
 
     def __init__(self, config: "StorageConfig") -> None:
         """Set the StorageConfig, sets up an empty Storage or loads existing data, ensures indexing,
@@ -37,6 +38,7 @@ class StorageInterface:
             AssertionError: If family_id 0 already exists, but does not refer to Family \"\".
         """
         self._config = config
+        self._band_projection = None
 
     # -> Set[function_id]
     def getCandidatesForMinHash(self, minhash: "MinHash") -> Set[int]:
@@ -128,8 +130,7 @@ class StorageInterface:
 
     def importFunctionEntry(self, function_entry: "FunctionEntry") -> Optional["FunctionEntry"]:
         """Import a function_entry to storage based on a previously exported FunctionEntry.
-        We assume that the family_id and sample_id was already remapped by the MinHashIndex, meaning that only the function_id needs to be adjusted
-        When importing similarity hashes (pichash, minhash, ...) make sure they are added to their respective lookup indices. 
+        We assume that the family_id and sample_id were already remapped by the MinHashIndex, meaning that only the function_id needs to be adjusted.
 
         Args:
             function_entry: the FunctionEntry to add to the storage
@@ -139,6 +140,19 @@ class StorageInterface:
         """
         raise NotImplementedError
 
+
+    def importFunctionEntries(self, function_entries: List["FunctionEntry"]) -> Optional[List["FunctionEntry"]]:
+        """Import multiple function_entries to storage based on previously exported FunctionEntry objects.
+        We assume that the family_id and sample_id were already remapped by the MinHashIndex, meaning that only the function_ids need to be adjusted.
+
+        Args:
+            function_entries: List of FunctionEntry objects to add to the storage
+
+        Returns:
+            A list of adjusted FunctionEntry objects if successful or None, if we failed somewhere in the procedure
+        """
+        raise NotImplementedError
+
     def deleteSample(self, sample_id: int) -> bool:
         """Remove a sample from the storage, also removes all functions of the sample.
         All minhashes will be removed from the bands.
@@ -219,22 +233,6 @@ class StorageInterface:
         """
         raise NotImplementedError
 
-    # (check if sample_id exists?)
-    def _addFunction(
-        self, sample_entry: "SampleEntry", smda_function: "SmdaFunction", minhash: Optional["MinHash"] = None
-    ) -> "FunctionEntry":
-        """Add a function (and optionally its MinHash) to storage, using the respective SampleEntry for reference.
-
-        Args:
-            sample_entry: The SampleEntry of smda_function
-            smda_function: The SmdaFunction to be added.
-            minhash: (Optional, defaults to None) The MinHash of the Function
-
-        Returns:
-            A FunctionEntry
-        """
-        raise NotImplementedError
-
     def getFunctionsBySampleId(self, sample_id: int) -> Optional[List["FunctionEntry"]]:
         """For a given sample_id, get all corresponding FunctionEntries.
 
@@ -246,11 +244,12 @@ class StorageInterface:
         """
         raise NotImplementedError
 
-    def getFunctionById(self, function_id: int) -> Optional["FunctionEntry"]:
+    def getFunctionById(self, function_id: int, with_xcfg=False) -> Optional["FunctionEntry"]:
         """Given a function_id, return the respective FunctionEntry or None, if function_id is not contained otherwise.
 
         Args:
             function_id: a function id
+            with_xcfg: include xcfg info (default: False)
 
         Returns:
             the respective FunctionEntry or None, if function_id is not contained otherwise
@@ -388,6 +387,18 @@ class StorageInterface:
         """
         raise NotImplementedError
 
+    def deleteFamily(self, family_id: int, keep_samples: Optional[str] = False) -> bool:
+        """Delete family if known and return boolean success state
+
+        Args:
+            family_id: The family_id of the family to delete
+            keep_samples: instead of deleting all samples and functions, reassign their family to "Unnamed (0)"
+
+        Returns:
+            True if successful
+        """
+        raise NotImplementedError
+
     def getFamilyIds(self) -> List[int]:
         """Returns a List of all family ids
 
@@ -471,14 +482,26 @@ class StorageInterface:
         """
         raise NotImplementedError
 
-    def getMatchesForPicHash(self, pichash: int) -> Set[Tuple[int, int]]:
-        """Get the set of all (sample_id, function_id) tuples for a given PicHash. If no function has the given pichash, the empty set is returned
+    def getMatchesForPicHash(self, pichash: int) -> Set[Tuple[int, int, int]]:
+        """Get the set of all (family_id, sample_id, function_id) tuples for a given PicHash. If no function has the given pichash, the empty set is returned
 
         Args:
             pichash: the pichash to look up
 
         Returns:
-            a set of (sample_id, function_id) tuples with the given pichash
+            a set of (family_id, sample_id, function_id) tuples with the given pichash
+        """
+        raise NotImplementedError
+
+
+    def getMatchesForPicBlockHash(self, picblockhash: int) -> Set[Tuple[int, int, int, int]]:
+        """Get the set of all (family_id, sample_id, function_id, offset) tuples for a given PicBlockHash. If no function has the given picblockhash, the empty set is returned
+
+        Args:
+            picblockhash: the picblockhash to look up
+
+        Returns:
+            a set of (family_id, sample_id, function_id, offset) tuples with the given picblockhash
         """
         raise NotImplementedError
 
@@ -533,26 +556,38 @@ class StorageInterface:
         """
         raise NotImplementedError
 
-    # -> Dict[BandIndex, BandHash]
-    def getBandHashesForMinHash(self, minhash: "MinHash") -> Dict[int, int]:
-        """Calculate band hashes for a given minhash, based on config parameters (STORAGE_BAND_SEED, STORAGE_BANDS)
-
+    def createBandhashProjection(self, minhash):
+        """Calculate a projection for index permutation based on a given minhash
         Args:
-            minhash: the MinHash for which the BandHashes will be calculated
-
+            minhash: the MinHash used as reference
         Returns:
-            a dict containing BandHashes by BandId
+            a dict containing signature indices used for bandhashing by band id
         """
-        band_hashes = {}
+        band_projection = {}
         random.seed(self._config.STORAGE_BAND_SEED)
         band_index = 0
-        minhash_data = minhash.getMinHashInt()
         for band_size, num_bands in self._config.STORAGE_BANDS.items():
             for _ in range(num_bands):
                 index_sequence = [index for index in range(len(minhash.getMinHashInt()))]
                 random.shuffle(index_sequence)
-                band_data = [minhash_data[i] for i in index_sequence[:band_size]]
-                hashed_band_data = MinHash.hashData(band_data, 0)
-                band_hashes[band_index] = hashed_band_data
+                band_projection[band_index] = index_sequence[:band_size]
                 band_index += 1
+        return band_projection
+
+    # -> Dict[BandIndex, BandHash]
+    def getBandHashesForMinHash(self, minhash: "MinHash") -> Dict[int, int]:
+        """Calculate band hashes for a given minhash, based on config parameters (STORAGE_BAND_SEED, STORAGE_BANDS)
+        Args:
+            minhash: the MinHash for which the BandHashes will be calculated
+        Returns:
+            a dict containing BandHashes by BandId
+        """
+        if self._band_projection is None:
+            self._band_projection = self.createBandhashProjection(minhash)
+        band_hashes = {}
+        minhash_data = minhash.getMinHashInt()
+        for band_index, permutation in self._band_projection.items():
+            band_data = [minhash_data[i] for i in permutation]
+            hashed_band_data = MinHash.hashData(band_data, 0)
+            band_hashes[band_index] = hashed_band_data
         return band_hashes
